﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading;
using System.Windows;

namespace CassiniSwitch.Model
{
	class CassiniInstaller
	{
		public bool IsVS2008Available { get; private set; }
		public bool IsVS2008DevServer { get; private set; }
		public bool IsVS2008Cassini { get; private set; }
		public bool IsVS2008X64 { get; private set; }
		public string VS2008Version { get; private set; }

		public bool IsVS2010Available { get; private set; }
		public bool IsVS2010DevServer { get; private set; }
		public bool IsVS2010Cassini { get; private set; }
		public bool IsVS2010X64 { get; private set; }
		public string VS2010Version { get; private set; }

		public bool IsVS2012Available { get; private set; }
		public bool IsVS2012DevServer { get; private set; }
		public bool IsVS2012Cassini { get; private set; }
		public bool IsVS2012X64 { get; private set; }
		public string VS2012Version { get; private set; }
		
		public CassiniInstaller()
		{
			UpdateData();
		}

		private void UpdateData()
		{
			IsVS2008Available = IsVSAvailable(2008);
			IsVS2008DevServer = IsDevServerInstalled(2008);
			IsVS2008Cassini = IsCassiniInstalled(2008);
			IsVS2008X64 = IsCassiniX64(2008);
			VS2008Version = GetInstalledVersion(2008);

			IsVS2010Available = IsVSAvailable(2010);
			IsVS2010DevServer = IsDevServerInstalled(2010);
			IsVS2010Cassini = IsCassiniInstalled(2010);
			IsVS2010X64 = IsCassiniX64(2010);
			VS2010Version = GetInstalledVersion(2010);

			IsVS2012Available = IsVSAvailable(2012);
			IsVS2012DevServer = IsDevServerInstalled(2012);
			IsVS2012Cassini = IsCassiniInstalled(2012);
			IsVS2012X64 = IsCassiniX64(2012);
			VS2012Version = GetInstalledVersion(2012);
		}

		public string GetInstallDirectory(int vsVersion)
		{
			string versionDirectory;
			switch (vsVersion)
			{
				case 0:
					versionDirectory = "";
					break;
				case 2008:
					versionDirectory = "9.0";
					break;
				case 2010:
					versionDirectory = "10.0";
					break;
				case 2012:
					versionDirectory = "11.0";
					break;
				default:
					throw new NotSupportedException("Unsupported VS version " + vsVersion);
			}

			string programFiles = Environment.GetFolderPath(Environment.SpecialFolder.CommonProgramFilesX86);
			if (string.IsNullOrEmpty(programFiles))
			{
				programFiles = Environment.GetFolderPath(Environment.SpecialFolder.CommonProgramFiles);
			}
			
			return Path.Combine(
				programFiles,
				"Microsoft Shared",
				"DevServer",
				versionDirectory);
		}

		/// <summary>
		/// Gets the full path of the WebDev executable file.
		/// </summary>
		/// <param name="netVersion">.NET version, multiplied with 10. Example: 20, 40, 45. Ignored for VS 2008.</param>
		/// <returns></returns>
		private string GetExecutablePath(int vsVersion, int netVersion)
		{
			return Path.Combine(GetInstallDirectory(vsVersion), "WebDev.WebServer" + (vsVersion > 2008 ? netVersion.ToString() : "") + ".exe");
		}

		public bool IsVSAvailable(int vsVersion)
		{
			string fileName = GetInstallDirectory(vsVersion);
			return Directory.Exists(fileName);
		}

		public bool IsDevServerInstalled(int vsVersion, int netVersion = 20)
		{
			string fileName = GetExecutablePath(vsVersion, netVersion);
			if (!File.Exists(fileName)) return false;
			var version = FileVersionInfo.GetVersionInfo(fileName);
			return !version.ProductName.ToLowerInvariant().Contains("cassini");
		}

		public bool IsCassiniInstalled(int vsVersion, int netVersion = 20)
		{
			string fileName = GetExecutablePath(vsVersion, netVersion);
			if (!File.Exists(fileName)) return false;
			var version = FileVersionInfo.GetVersionInfo(fileName);
			return version.ProductName.ToLowerInvariant().Contains("cassini");
		}

		public bool IsCassiniX64(int vsVersion, int netVersion = 20)
		{
			string fileName = GetExecutablePath(vsVersion, netVersion);
			if (!File.Exists(fileName)) return false;
			var an = AssemblyName.GetAssemblyName(fileName);
			return an.ProcessorArchitecture == ProcessorArchitecture.Amd64;
		}

		public string GetInstalledVersion(int vsVersion)
		{
			string fileName = GetExecutablePath(vsVersion, 20);
			if (!File.Exists(fileName)) return null;
			var version = FileVersionInfo.GetVersionInfo(fileName);
			if (version.ProductName.ToLowerInvariant().Contains("cassini"))
			{
				return version.ProductVersion;
			}
			else
			{
				return null;
			}
		}

		public void UninstallCassini(int vsVersion)
		{
			string instDir = GetInstallDirectory(vsVersion);
			string backupDir = instDir + "_orig";

			if (!IsDevServerInstalled(vsVersion))
			{
				if (Directory.Exists(backupDir))
				{
					// Delete existing installation directory
					if (Directory.Exists(instDir))
					{
						try
						{
							Directory.Delete(instDir, true);
						}
						catch (Exception ex)
						{
							ReportError("The Cassini installation directory could not be deleted.", ex);
							return;
						}
					}
					// Restore DevServer files from backup
					try
					{
						Directory.Move(backupDir, instDir);
					}
					catch (Exception ex)
					{
						ReportError("The DevServer backup directory could not be restored.", ex);
						return;
					}
				}
				else
				{
					ReportError("The DevServer backup directory does not exist.", null);
					return;
				}
			}

			// Update local properties
			UpdateData();
		}

		public void InstallCassini(int vsVersion, bool x64)
		{
			string instDir = GetInstallDirectory(vsVersion);
			string backupDir = instDir + "_orig";

			if (Directory.Exists(instDir) &&
				IsDevServerInstalled(vsVersion))
			{
				if (!Directory.Exists(backupDir))
				{
					// Create backup of DevServer files: rename directory
					try
					{
						Directory.Move(instDir, backupDir);
					}
					catch (Exception ex)
					{
						ReportError("The DevServer installation directory could not be moved to a backup name.", ex);
						return;
					}
				}
				else
				{
					ReportError("The DevServer backup directory already exists but was not expected to.\n" +
						"Please check the directories and delete the backup directory \"" + backupDir + "\".", null);
					return;
				}
			}

			// Create installation directory, if it doesn't exist
			try
			{
				Directory.CreateDirectory(instDir);
			}
			catch (Exception ex)
			{
				ReportError("The installation directory could not be created.", ex);
				return;
			}
			// Write new Cassini application files
			string platform = x64 ? "x64" : "x86";
			try
			{
				if (vsVersion > 2008)
				{
					WriteResourceFile("appfiles." + platform, "WebDev.WebServer20.exe", instDir);
					WriteResourceFile("appfiles." + platform, "WebDev.WebServer20.exe.config", instDir);
					WriteResourceFile("appfiles." + platform, "WebDev.WebServer40.exe", instDir);
					WriteResourceFile("appfiles." + platform, "WebDev.WebServer40.exe.config", instDir);
				}
				else
				{
					WriteResourceFile("appfiles." + platform, "WebDev.WebServer20.exe", instDir, "WebDev.WebServer.exe");
					WriteResourceFile("appfiles." + platform, "WebDev.WebServer20.exe.config", instDir, "WebDev.WebServer.exe.config");
				}
			}
			catch (Exception ex)
			{
				ReportError("The Cassini application files could not be saved in the installation directory.", ex);
				return;
			}

			// Update local properties
			UpdateData();
		}

		public void ReportError(string msg, Exception ex)
		{
			MessageBox.Show(
				msg + (ex != null ? "\n" + ex.Message : ""),
				"Installation error",
				MessageBoxButton.OK,
				MessageBoxImage.Error);
		}

		private void WriteResourceFile(string resourcePath, string resourceName, string targetDir, string targetName = null)
		{
			if (targetName == null)
				targetName = resourceName;
			string targetFile = Path.Combine(targetDir, targetName);
			foreach (Process p in Process.GetProcessesByName(Path.GetFileNameWithoutExtension(targetName)))
			{
				if (PathEquals(p.MainModule.FileName, targetFile))
				{
					if (MessageBox.Show(
						"The Cassini application to be installed is currently running. Terminate and continue?",
						"Process running",
						MessageBoxButton.YesNo,
						MessageBoxImage.Warning) != MessageBoxResult.Yes)
					{
						return;
					}
					try
					{
						p.CloseMainWindow();
						if (!p.WaitForExit(500))
						{
							p.Kill();
						}
					}
					catch (InvalidOperationException)
					{
						// Process was probably already terminated by the user.
					}
					Thread.Sleep(250);
					break;
				}
			}

			using (Stream inStream = Assembly.GetExecutingAssembly().GetManifestResourceStream("CassiniSwitch." + resourcePath + "." + resourceName))
			using (FileStream outStream = new FileStream(targetFile, FileMode.Create))
			{
				inStream.CopyTo(outStream);
			}
		}

		/// <summary>
		/// Determines whether two paths are equal. This does not consider file system links of any
		/// kind or UNC paths.
		/// </summary>
		/// <param name="path1"></param>
		/// <param name="path2"></param>
		/// <returns></returns>
		private static bool PathEquals(string path1, string path2)
		{
			if (string.IsNullOrWhiteSpace(path1)) return false;
			if (string.IsNullOrWhiteSpace(path2)) return false;

			return String.Compare(
				Path.GetFullPath(path1.Trim()).TrimEnd('\\'),
				Path.GetFullPath(path2.Trim()).TrimEnd('\\'),
				Environment.OSVersion.Platform == PlatformID.Unix ?
					StringComparison.InvariantCulture :
					StringComparison.InvariantCultureIgnoreCase) == 0;
		}

		public static string AssemblyFileVersion
		{
			get
			{
				object[] customAttributes = Assembly.GetEntryAssembly().GetCustomAttributes(typeof(AssemblyFileVersionAttribute), false);
				if (customAttributes != null && customAttributes.Length > 0)
				{
					return ((AssemblyFileVersionAttribute) customAttributes[0]).Version;
				}
				return null;
			}
		}

		public static string AssemblyCopyright
		{
			get
			{
				object[] customAttributes = Assembly.GetEntryAssembly().GetCustomAttributes(typeof(AssemblyCopyrightAttribute), false);
				if (customAttributes != null && customAttributes.Length > 0)
				{
					return ((AssemblyCopyrightAttribute) customAttributes[0]).Copyright;
				}
				return null;
			}
		}
	}
}
